import torch
import torch.nn.functional as F
import torch.utils.data.sampler as samplers
from torch.utils.data import Dataset, DataLoader
from torch.nn.utils.rnn import pad_sequence
import numpy as np
import math
from pathlib import Path
import json

from tacotron import load_cmudict, text_to_id


class SortedSampler(samplers.Sampler):
    """
    Adapted from https://github.com/PetrochukM/PyTorch-NLP/blob/master/torchnlp/samplers/sorted_sampler.py
    Copyright (c) James Bradbury and Soumith Chintala 2016,
    All rights reserved.
    """

    def __init__(self, data, sort_key):
        super().__init__(data)
        self.data = data
        self.sort_key = sort_key
        zip_ = [(i, self.sort_key(row)) for i, row in enumerate(self.data)]
        zip_ = sorted(zip_, key=lambda r: r[1], reverse=True)
        self.sorted_indexes = [item[0] for item in zip_]

    def __iter__(self):
        return iter(self.sorted_indexes)

    def __len__(self):
        return len(self.data)


class BucketBatchSampler(samplers.BatchSampler):
    """
    Adapted from https://github.com/PetrochukM/PyTorch-NLP/blob/master/torchnlp/samplers/bucket_batch_sampler.py
    Copyright (c) James Bradbury and Soumith Chintala 2016,
    All rights reserved.
    """

    def __init__(
        self,
        sampler,
        batch_size,
        drop_last,
        sort_key,
        bucket_size_multiplier=100,
    ):
        super().__init__(sampler, batch_size, drop_last)
        self.sort_key = sort_key
        self.bucket_sampler = samplers.BatchSampler(
            sampler, min(batch_size * bucket_size_multiplier, len(sampler)), False
        )

    def __iter__(self):
        for bucket in self.bucket_sampler:
            sorted_sampler = SortedSampler(bucket, self.sort_key)
            for batch in samplers.SubsetRandomSampler(
                list(
                    samplers.BatchSampler(
                        sorted_sampler, self.batch_size, self.drop_last
                    )
                )
            ):
                yield [bucket[i] for i in batch]

    def __len__(self):
        if self.drop_last:
            return len(self.sampler) // self.batch_size
        else:
            return math.ceil(len(self.sampler) / self.batch_size)


class TTSDataset(Dataset):
    def __init__(self, root, text_path):
        self.root = Path(root)

        with open(self.root / "train.json") as file:
            metadata = json.load(file)
            self.metadata = [Path(path) for _, path in metadata]

        with open(self.root / "lengths.json") as file:
            lengths = json.load(file)

            self.lengths = list()
            for path in self.metadata:
                self.lengths.append(lengths[path.stem])

        self.index_longest_mel = np.argmax(self.lengths)

        self.cmudict = load_cmudict()

        keys = {path.stem for path in self.metadata}
        with open(text_path) as file:
            self.text = {}
            for line in file:
                key, _, transcript = line.strip().split("|")
                if key in keys:
                    self.text[key] = transcript

    def sort_key(self, index):
        return self.lengths[index]

    def __len__(self):
        return len(self.metadata)

    def __getitem__(self, index):
        path = self.root / self.metadata[index]

        mel = np.load(path.with_suffix(".mel.npy"))

        text = text_to_id(self.text[path.stem], self.cmudict)

        return (
            torch.Tensor(mel).transpose_(0, 1),
            torch.LongTensor(symbols),
            index == self.index_longest_mel,
        )


def pad_collate(batch):
    mels, texts, attn_flag = zip(*batch)
    mels = list(mels)
    texts = list(texts)

    # TODO: handle general reduction factor
    if len(mels[0]) % 2 != 0:
        mels[0] = F.pad(mels[0], (0, 0, 0, 1))

    mel_lengths = [len(mel) for mel in mels]
    text_lengths = [len(text) for text in texts]

    mels = pad_sequence(mels, batch_first=True)
    texts = pad_sequence(texts, batch_first=True, padding_value=symbol_to_id["_"])

    attn_flag = [i for i, flag in enumerate(attn_flag) if flag]

    return mels.transpose_(1, 2), texts, mel_lengths, text_lengths, attn_flag



""" adapted from https://github.com/keithito/tacotron """

import re
from itertools import islice
import importlib_resources

SYMBOLS = [
    "_",  # padding
    "~",  # end of sentence
    " ",
    "!",
    ",",
    ".",
    "?",
    "AA",
    "AA0",
    "AA1",
    "AA2",
    "AE",
    "AE0",
    "AE1",
    "AE2",
    "AH",
    "AH0",
    "AH1",
    "AH2",
    "AO",
    "AO0",
    "AO1",
    "AO2",
    "AW",
    "AW0",
    "AW1",
    "AW2",
    "AY",
    "AY0",
    "AY1",
    "AY2",
    "B",
    "CH",
    "D",
    "DH",
    "EH",
    "EH0",
    "EH1",
    "EH2",
    "ER",
    "ER0",
    "ER1",
    "ER2",
    "EY",
    "EY0",
    "EY1",
    "EY2",
    "F",
    "G",
    "HH",
    "IH",
    "IH0",
    "IH1",
    "IH2",
    "IY",
    "IY0",
    "IY1",
    "IY2",
    "JH",
    "K",
    "L",
    "M",
    "N",
    "NG",
    "OW",
    "OW0",
    "OW1",
    "OW2",
    "OY",
    "OY0",
    "OY1",
    "OY2",
    "P",
    "R",
    "S",
    "SH",
    "T",
    "TH",
    "UH",
    "UH0",
    "UH1",
    "UH2",
    "UW",
    "UW0",
    "UW1",
    "UW2",
    "V",
    "W",
    "Y",
    "Z",
    "ZH",
]

symbol_to_id = {s: i for i, s in enumerate(SYMBOLS)}
id_to_symbol = {i: s for i, s in enumerate(SYMBOLS)}

abbreviations = [
    (re.compile("\\b%s\\." % x[0], re.IGNORECASE), x[1].upper())
    for x in [
        ("mrs", "missis"),
        ("mr", "mister"),
        ("dr", "doctor"),
        ("st", "saint"),
        ("co", "company"),
        ("jr", "junior"),
        ("maj", "major"),
        ("gen", "general"),
        ("drs", "doctors"),
        ("rev", "reverend"),
        ("lt", "lieutenant"),
        ("hon", "honorable"),
        ("sgt", "sergeant"),
        ("capt", "captain"),
        ("esq", "esquire"),
        ("ltd", "limited"),
        ("col", "colonel"),
        ("ft", "fort"),
        ("etc", "etcetera"),
    ]
]


def expand_abbreviations(text):
    for regex, replacement in abbreviations:
        text = re.sub(regex, replacement, text)
    return text


alternate_pronunciation_regex = re.compile(r"(?<=\w)\((\d)\)")
parentheses_regex = re.compile(r"(?<=[.,!?] )[\(\[]|[\)\]](?=[.,!?])|^[\(\[]|[\)\]]$")
dash_regex = re.compile(r"(?<=[.,!?] )-- ")


def replace_symbols(text, lang="en"):
    # replace semi-colons and colons with commas
    text = text.replace(";", ",")
    text = text.replace(":", ",")

    # replace dashes with commas
    text = re.sub(dash_regex, "", text)
    text = text.replace(" --", ",")
    text = text.replace(" - ", ", ")

    # split hyphenated words
    text = text.replace("-", " ")

    # use {#} to indicate alternate pronunciations
    text = re.sub(alternate_pronunciation_regex, r"{\1}", text)

    # replace parentheses with commas
    text = re.sub(parentheses_regex, "", text)
    text = text.replace(")", ",")
    text = text.replace(" (", ", ")
    text = text.replace("]", ",")
    text = text.replace(" [", ", ")
    return text


def clean(text):
    text = text.upper()
    text = expand_abbreviations(text)
    text = replace_symbols(text)
    return text


tokenizer_regex = re.compile(r"[\w\{\}']+|[.,!?]")


def tokenize(text):
    return re.findall(tokenizer_regex, text)


def load_cmudict():
    cmudict = dict()
    dict_ref = importlib_resources.files("tacotron.dictionary").joinpath(
        "cmudict-0.7b.txt"
    )
    with open(dict_ref, encoding="ISO-8859-1") as file:
        for line in islice(file, 126, 133905):
            word, pronunciation = line.strip().split("  ")
            word = re.sub(alternate_pronunciation_regex, r"{\1}", word)
            cmudict[word] = pronunciation
    return cmudict


def parse_text(text, cmudict):
    text = clean(text)
    symbols, out_of_vocabulary = list(), set()
    for word in tokenize(text):
        word_symbols = cmudict.get(word, word if word in ".,!?" else None)
        if word_symbols is not None:
            symbols.append(word_symbols.split(" "))
        else:
            out_of_vocabulary.add(word)

    if out_of_vocabulary:
        raise KeyError(
            f"Please add the following words to the pronunciation dictionary: {', '.join(out_of_vocabulary)}"
        )

    symbols = [symbol for word in symbols for symbol in (word, [" "])]
    symbols.append(["~"])
    return symbols


def symbols_to_id(symbols):
    return [symbol_to_id[symbol] for word in symbols for symbol in word]


def text_to_id(text, cmudict):
    symbols = parse_text(text, cmudict)
    return symbols_to_id(symbols)


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from scipy.stats import betabinom

import importlib_resources
from omegaconf import OmegaConf


class PreNet(nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        output_size,
        dropout=0.5,
        fixed=False,
    ):
        super().__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.fc2 = nn.Linear(hidden_size, output_size)
        self.p = dropout
        self.fixed = fixed

    def forward(self, x):
        x = self.fc1(x)
        x = F.relu(x)
        x = F.dropout(x, self.p, training=self.training or self.fixed)
        x = self.fc2(x)
        x = F.relu(x)
        x = F.dropout(x, self.p, training=self.training or self.fixed)
        return x


class BatchNormConv(nn.Module):
    def __init__(self, input_channels, output_channels, kernel_size, relu=True):
        super().__init__()
        self.conv = nn.Conv1d(
            input_channels,
            output_channels,
            kernel_size,
            stride=1,
            padding=kernel_size // 2,
            bias=False,
        )
        self.bnorm = nn.BatchNorm1d(output_channels)
        self.relu = relu

    def forward(self, x):
        x = self.conv(x)
        x = F.relu(x) if self.relu is True else x
        return self.bnorm(x)


class HighwayNetwork(nn.Module):
    def __init__(self, size):
        super().__init__()
        self.linear1 = nn.Linear(size, size)
        self.linear2 = nn.Linear(size, size)
        nn.init.zeros_(self.linear1.bias)

    def forward(self, x):
        x1 = self.linear1(x)
        x2 = self.linear2(x)
        g = torch.sigmoid(x2)
        return g * F.relu(x1) + (1.0 - g) * x


class CBHG(nn.Module):
    def __init__(
        self,
        K,
        input_channels,
        channels,
        projection_channels,
        n_highways,
        highway_size,
        rnn_size,
    ):
        super().__init__()

        self.conv_bank = nn.ModuleList(
            [
                BatchNormConv(input_channels, channels, kernel_size)
                for kernel_size in range(1, K + 1)
            ]
        )
        self.max_pool = nn.MaxPool1d(kernel_size=2, stride=1, padding=1)

        self.conv_projections = nn.Sequential(
            BatchNormConv(K * channels, projection_channels, 3),
            BatchNormConv(projection_channels, input_channels, 3, relu=False),
        )

        self.project = (
            nn.Linear(input_channels, highway_size, bias=False)
            if input_channels != highway_size
            else None
        )

        self.highway = nn.Sequential(
            *[HighwayNetwork(highway_size) for _ in range(n_highways)]
        )

        self.rnn = nn.GRU(highway_size, rnn_size, batch_first=True, bidirectional=True)

    def forward(self, x):
        T = x.size(-1)
        residual = x

        stack = []
        for conv in self.conv_bank:
            stack.append(conv(x)[:, :, :T])
        x = torch.cat(stack, dim=1)

        x = self.max_pool(x)

        x = self.conv_projections(x[:, :, :T])

        x = x + residual
        x = x.transpose(1, 2)

        if self.project is not None:
            x = self.project(x)

        x = self.highway(x)

        x, _ = self.rnn(x)
        return x


class Encoder(nn.Module):
    def __init__(self, n_symbols, embedding_dim, prenet, cbhg):
        super().__init__()
        self.embedding = nn.Embedding(n_symbols, embedding_dim)
        self.pre_net = PreNet(**prenet)
        self.cbhg = CBHG(**cbhg)

    def forward(self, x):
        x = self.embedding(x)
        x = self.pre_net(x)
        x = self.cbhg(x.transpose(1, 2))
        return x


class DynamicConvolutionAttention(nn.Module):
    def __init__(
        self,
        attn_rnn_size,
        hidden_size,
        static_channels,
        static_kernel_size,
        dynamic_channels,
        dynamic_kernel_size,
        prior_length,
        alpha,
        beta,
    ):
        super(DynamicConvolutionAttention, self).__init__()

        self.prior_length = prior_length
        self.dynamic_channels = dynamic_channels
        self.dynamic_kernel_size = dynamic_kernel_size

        P = betabinom.pmf(np.arange(prior_length), prior_length - 1, alpha, beta)

        self.register_buffer("P", torch.FloatTensor(P).flip(0))
        self.W = nn.Linear(attn_rnn_size, hidden_size)
        self.V = nn.Linear(
            hidden_size, dynamic_channels * dynamic_kernel_size, bias=False
        )
        self.F = nn.Conv1d(
            1,
            static_channels,
            static_kernel_size,
            padding=(static_kernel_size - 1) // 2,
            bias=False,
        )
        self.U = nn.Linear(static_channels, hidden_size, bias=False)
        self.T = nn.Linear(dynamic_channels, hidden_size)
        self.v = nn.Linear(hidden_size, 1, bias=False)

    def forward(self, s, alpha):
        p = F.conv1d(
            F.pad(alpha.unsqueeze(1), (self.prior_length - 1, 0)), self.P.view(1, 1, -1)
        )
        p = torch.log(p.clamp_min_(1e-6)).squeeze(1)

        G = self.V(torch.tanh(self.W(s)))
        g = F.conv1d(
            alpha.unsqueeze(0),
            G.view(-1, 1, self.dynamic_kernel_size),
            padding=(self.dynamic_kernel_size - 1) // 2,
            groups=s.size(0),
        )
        g = g.view(s.size(0), self.dynamic_channels, -1).transpose(1, 2)

        f = self.F(alpha.unsqueeze(1)).transpose(1, 2)

        e = self.v(torch.tanh(self.U(f) + self.T(g))).squeeze(-1) + p

        return F.softmax(e, dim=-1)


def zoneout(prev, current, p=0.1):
    mask = torch.empty_like(prev).bernoulli_(p)
    return mask * prev + (1 - mask) * current


class DecoderCell(nn.Module):
    def __init__(
        self,
        prenet,
        attention,
        input_size,
        n_mels,
        attn_rnn_size,
        decoder_rnn_size,
        reduction_factor,
        zoneout_prob,
    ):
        super(DecoderCell, self).__init__()
        self.zoneout_prob = zoneout_prob

        self.prenet = PreNet(**prenet)
        self.dca = DynamicConvolutionAttention(**attention)
        self.attn_rnn = nn.LSTMCell(2 * input_size + prenet.output_size, attn_rnn_size)
        self.linear = nn.Linear(2 * input_size + decoder_rnn_size, decoder_rnn_size)
        self.rnn1 = nn.LSTMCell(decoder_rnn_size, decoder_rnn_size)
        self.rnn2 = nn.LSTMCell(decoder_rnn_size, decoder_rnn_size)
        self.proj = nn.Linear(decoder_rnn_size, n_mels * reduction_factor, bias=False)

    def forward(self, h, y, alpha, c, attn_hx, rnn1_hx, rnn2_hx):
        B, N = y.size()

        y = self.prenet(y)
        attn_h, attn_c = self.attn_rnn(torch.cat((c, y), dim=-1), attn_hx)
        if self.training:
            attn_h = zoneout(attn_hx[0], attn_h, p=self.zoneout_prob)

        alpha = self.dca(attn_h, alpha)

        c = torch.matmul(alpha.unsqueeze(1), h).squeeze(1)

        x = self.linear(torch.cat((c, attn_h), dim=-1))

        rnn1_h, rnn1_c = self.rnn1(x, rnn1_hx)
        if self.training:
            rnn1_h = zoneout(rnn1_hx[0], rnn1_h, p=self.zoneout_prob)
        x = x + rnn1_h

        rnn2_h, rnn2_c = self.rnn2(x, rnn2_hx)
        if self.training:
            rnn2_h = zoneout(rnn2_hx[0], rnn2_h, p=self.zoneout_prob)
        x = x + rnn2_h

        y = self.proj(x).view(B, N, 2)
        return y, alpha, c, (attn_h, attn_c), (rnn1_h, rnn1_c), (rnn2_h, rnn2_c)


class Tacotron(nn.Module):
    def __init__(self, encoder, decoder):
        super().__init__()
        self.input_size = 2 * decoder.input_size
        self.attn_rnn_size = decoder.attn_rnn_size
        self.decoder_rnn_size = decoder.decoder_rnn_size
        self.n_mels = decoder.n_mels
        self.reduction_factor = decoder.reduction_factor

        self.dictionary = None

        self.encoder = Encoder(**encoder)
        self.decoder_cell = DecoderCell(**decoder)

    @classmethod
    def from_pretrained(cls, url, cfg_path=None):
        r"""
        Loads the Torch serialized object at the given URL (uses torch.hub.load_state_dict_from_url).

        Parameters:
            url (string): URL of the weights to download
            cfg_path (Path): path to config file. Defaults to tacotron/config/config.yaml
        """
        cfg_ref = (
            importlib_resources.files("tacotron.config").joinpath("config.yaml")
            if cfg_path is None
            else cfg_path
        )
        with cfg_ref.open() as file:
            cfg = OmegaConf.load(file)
        checkpoint = torch.hub.load_state_dict_from_url(url)
        model = cls(**cfg.model)
        model.load_state_dict(checkpoint["tacotron"])
        model.eval()
        return model

    def forward(self, x, mels):
        B, N, T = mels.size()

        h = self.encoder(x)

        alpha = F.one_hot(
            torch.zeros(B, dtype=torch.long, device=x.device), h.size(1)
        ).float()
        c = torch.zeros(B, self.input_size, device=x.device)

        attn_hx = (
            torch.zeros(B, self.attn_rnn_size, device=x.device),
            torch.zeros(B, self.attn_rnn_size, device=x.device),
        )

        rnn1_hx = (
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
        )

        rnn2_hx = (
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
        )

        go_frame = torch.zeros(B, N, device=x.device)

        ys, alphas = list(), list()
        for t in range(0, T, self.reduction_factor):
            y = mels[:, :, t - 1] if t > 0 else go_frame
            y, alpha, c, attn_hx, rnn1_hx, rnn2_hx = self.decoder_cell(
                h, y, alpha, c, attn_hx, rnn1_hx, rnn2_hx
            )
            ys.append(y)
            alphas.append(alpha)

        ys = torch.cat(ys, dim=-1)
        alphas = torch.stack(alphas, dim=2)
        return ys, alphas

    def generate(self, x, max_length=10000, stop_threshold=-0.2):
        h = self.encoder(x)
        B, T, _ = h.size()

        alpha = F.one_hot(torch.zeros(B, dtype=torch.long, device=x.device), T).float()
        c = torch.zeros(B, self.input_size, device=x.device)

        attn_hx = (
            torch.zeros(B, self.attn_rnn_size, device=x.device),
            torch.zeros(B, self.attn_rnn_size, device=x.device),
        )

        rnn1_hx = (
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
        )

        rnn2_hx = (
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
            torch.zeros(B, self.decoder_rnn_size, device=x.device),
        )

        go_frame = torch.zeros(B, self.n_mels, device=x.device)

        k = 0
        ys, alphas = list(), list()
        for t in range(0, max_length, self.reduction_factor):
            y = ys[-1][:, :, -1] if t > 0 else go_frame
            y, alpha, c, attn_hx, rnn1_hx, rnn2_hx = self.decoder_cell(
                h, y, alpha, c, attn_hx, rnn1_hx, rnn2_hx
            )
            if torch.all(y[:, :, -1] > stop_threshold):
                break
            ys.append(y)
            alphas.append(alpha)

        ys = torch.cat(ys, dim=-1)
        alphas = torch.stack(alphas, dim=2)
        return ys, alphas
